#' NT Spatial Weights Matrix for Unbalanced TSCS Samples.
#'
#' \code{make_ntspmat_ch} generates a nearest neighbor spatial
#' weights matrix for an unbalanced TSCS sample used to
#' estimate a linear regression model.
#'
#' @param lmobj An object created by the \code{lm} function.
#' @param ci Name of a variable that identifies countries.
#' @param yi Name of a variable that identifies years.
#' @param k A value that determines the number of nearest
#' neighbors.
#' @return The output will be a matrix.
#' @example
#' results <- make_ntspmat_ch(ols, country_name, year, 5)
make_ntspmat_ch <- function(lmobj,ci,yi,k) {

  # Identify and subset the sample from the regression, using the estimated disturbances.

  call <- match.call()
  esample<-rownames(as.matrix(lmobj[["residuals"]]))
  df <- eval(lmobj[["call"]][["data"]])
  df <- df[esample,]

  attach(df)
  c<-unique(ci)
  y<-sort(as.factor(unique(yi)),decreasing=FALSE)
  yl<-length(y)
  detach(df)

  df2<-as_tibble(df) %>%
    arrange(desc(-as.numeric(eval(parse(text=noquote(paste(call$yi))))))) %>%
    group_by(as.factor(eval(parse(text=noquote(paste(call$ci)))))) %>%
    slice(1) %>%
    ungroup()

  suppressWarnings(cshp.dat <- cshp())
  cshp.dat <- cshp.dat[cshp.dat$COWSYEAR < as.numeric(as.character(y[yl]))+1,]
  cow <- cbind(paste(cshp.dat@data[["CNTRY_NAME"]]),cshp.dat@data[["COWCODE"]],cshp.dat@data[["COWSYEAR"]]) %>%
  as.data.frame()
  names(cow)[1:3]<- c("CNTRY_NAME","COWCODE","COWSYEAR")
  cow<-cow[!(cow$COWCODE==-1),]

  cow2<-as_tibble(cow) %>%
    arrange(desc(-as.numeric(COWSYEAR))) %>%
    group_by(CNTRY_NAME) %>%
    slice(1) %>%
    ungroup()

  cow2$COWSYEAR<-as.numeric(as.character(cow2$COWSYEAR))
  cow2$COWSYEAR = cow2$COWSYEAR + (cow2$CNTRY_NAME == "Macedonia")

  o <- match(df2$`as.factor(eval(parse(text = noquote(paste(call$ci)))))`,cow2$CNTRY_NAME)
  cow2 <- cow2[o,]
  df2cow2 <- cbind(df2,cow2)
  df2cow2$diff<- df2cow2$COWSYEAR - as.numeric(as.character(eval(parse(text=noquote(paste("df2cow2$",call$yi,sep=""))))))
  df2cow2$diff2 <- ifelse(df2cow2$diff > 0, 1, df2cow2$diff)
  df2cow3 <- na.omit(df2cow2[df2cow2$diff2==1,])
  df2cow2 <- df2cow2[!(complete.cases(df2cow2$diff2)),]
  df2cow3 <- rbind(df2cow3,df2cow2)


  df2cow3<-df2cow3[,c(paste(call$ci),paste(call$yi),"CNTRY_NAME","COWSYEAR")]
  colnames(df2cow3)<-c("Data Country Name", "Data Start Year","COW Country Name","COW Start Year")
  df2cow3$"Data Start Year"<- as.character(df2cow3$"Data Start Year")
  df2cow3$"COW Start Year"<- as.character(df2cow3$"COW Start Year")
  rownames(df2cow3) <- c()

if(nrow(df2cow3) > 0) {
  stargazer(df2cow3,type="text",summary=FALSE)
  stop('Some of your Countries are not Matched.')
}


  o2 <- match(eval(parse(text=noquote(paste("df$",call$ci,sep="")))),cow2$CNTRY_NAME)
  cow2<-cow2[o2,]
  df.m <- cbind(df,cow2)
  df.m$coln <- paste(df.m$COWCODE,df.m$year,sep=".")

  # Necessary if Borders and Countries are Changing Over Time
  dmat <- suppressWarnings(as.data.frame(distmatrix(as.Date(y[1],"%Y"), type="capdist", useGW=F)))
  # Create first block diagonal
  dta <-df.m[which(df.m$year==as.numeric(as.character(y[1]))), ]
  cs <- na.omit(unique(dta$COWCODE))
  csl <- length(cs)
  dmat<-dmat[as.character(cs), as.character(cs)]

  dcol <- ncol(dmat)
  col2 <- dcol*dcol
  coln <- paste(colnames(dmat),y[1],sep=".")

  dsort <- t(apply(dmat,2,sort))
  dsort[,c(1:k+1)]<-1
  dsort[,seq(k+2,dcol)] <- 0
  dsort<-matrix(t(dsort), nrow=1, ncol=col2)

  dorder <- t(apply(dmat,2,order))
  seq <- seq(0,dcol-1)
  mmult <- diag(seq)
  mmult2 <- matrix(dcol,nrow=dcol,ncol=dcol)
  adj <- t(mmult2%*%mmult)
  dorder2 <- t(dorder + adj) %>%
    matrix(nrow=1, ncol=col2)

  seq <- t(as.matrix(seq(1,col2)))
  comb <- rbind(dorder2,seq)
  dorder3 <- t(comb[,order(comb[1,])]) %>%
    as.data.frame() %>%
    select(-1)

  nnmat <- t(as.matrix(dsort[t(dorder3)])) %>%
    matrix(nrow = dcol, byrow = TRUE)
  colnames(nnmat)<-coln
  rownames(nnmat)<-coln

  cyi <- as.data.frame(coln)
  nnmat_b <- nnmat

  print(noquote(as.character(y[1])))
  print(noquote(as.character(cs)))

  #Diagonally Concatenate Remaining Block Diagonal Matrices for Sparse NTxNT Matrix

  for (i in 2:yl) {

    dmat <- suppressWarnings(as.data.frame(distmatrix(as.Date(y[i],"%Y"), type="capdist", useGW=F)))
    dta <-df.m[which(df.m$year==as.numeric(as.character(y[i]))), ]
    cs <- as.character(na.omit(unique(dta$COWCODE))) #Need to See This
    csl <- length(cs)
    dmat<-dmat[as.character(cs), as.character(cs)]
    col <- ncol(dmat)
    col2 <- col*col
    coln <- paste(colnames(dmat),y[i],sep=".")

    dsort <- t(apply(dmat,2,sort))
    dsort[,c(1:k+1)]<-1
    dsort[,seq(k+2,col)] <- 0
    dsort<-matrix(t(dsort), nrow=1, ncol=col2)

    dorder <- t(apply(dmat,2,order))
    seq <- seq(0,col-1)
    mmult <- diag(seq)
    mmult2 <- matrix(col,nrow=col,ncol=col)
    adj <- t(mmult2%*%mmult)
    dorder2 <- t(dorder + adj) %>%
      matrix(nrow=1, ncol=col2)

    seq <- t(as.matrix(seq(1,col2)))
    comb <- rbind(dorder2,seq)
    dorder3 <- t(comb[,order(comb[1,])]) %>%
      as.data.frame() %>%
      select(-1)

    nnmat <- t(as.matrix(dsort[t(dorder3)])) %>%
      matrix(nrow = col, byrow = TRUE)
    colnames(nnmat)<-coln
    rownames(nnmat)<-coln
    cyi <- rbind(as.data.frame(cyi),as.data.frame(coln))

    nnmat_b <- bdiag(nnmat_b,nnmat)
    print(noquote(as.character(y[i])))
    print(noquote(cs))
    print("All of your Countries are Matched.", quote=FALSE)

  }

  nnmat <- as.matrix(nnmat_b)
  return(list(df.m,nnmat,cyi))
}

ntspreg <- function(lmobj,wm) {

  weights <- mat2listw(wm, row.names=NULL, style="W")
  nblist <-weights$neighbours
  listw <-nb2listw(nblist)

  formula <- lmobj[["call"]][["formula"]]
  esample<-rownames(as.matrix(lmobj[["residuals"]]))
  df <- eval(lmobj[["call"]][["data"]])
  df <- df[esample,]

  lag = spatialreg::lagsarlm(formula, data=df, listw=listw ,method="eigen",zero.policy=TRUE, tol.solve=1.0e-11)
  return(lag)
}

ntsperr <- function(lmobj,wm) {

  weights <- mat2listw(wm, row.names=NULL, style="W")
  nblist <-weights$neighbours
  listw <-nb2listw(nblist)

  formula <- lmobj[["call"]][["formula"]]
  esample<-rownames(as.matrix(lmobj[["residuals"]]))
  df <- eval(lmobj[["call"]][["data"]])
  df <- df[esample,]

  err = spatialreg::errorsarlm(formula, data=df, listw=listw ,method="eigen",zero.policy=TRUE, tol.solve=1.0e-11)
  return(err)
}
